#include "type.h"
#include "missile.h"
#include "asteroids.h"
#include "vect.h"
#include "rand.h"
#include "param.h"
#include "soundeff.h"

/*
 * The following used for flushes that are not needed except for on the
 * APA16.  Without the extra flushes, the APA16 graphics buffer overflows.
 * I hope this nonsense works...
 */

#define OPTIONAL_FLUSH() plot_Flush()

private SHAPE *Shapes[ASTSIZES][ASTROTATIONS];

#if defined(SUN) || defined(HP)
#define ADJASTVEL ((int) ((double) MAXASTVEL * 1.13 + 0.5))
#endif

#ifndef ADJASTVEL
#define ADJASTVEL MAXASTVEL
#endif

/*
 * The APA16 is speedy enough for a fairly complex asteroid.
 *
 * Sun asteroids should perhaps be simplified (less segments).
 * Sun asteroids are 13% bigger to make up for the increased
 * screen resolution.
 */

#if defined(SUN) || defined(HP)
private VECT asteroid[] = {
    {2,20},
    {11,25},
    {23,6},
    {19,-5},
    {8,-7},
    {18,-11},
    {14,-19},
    {-8,-21},
    {-19,-15},
    {-24,3},
    {-16,19},
    {-9,12},
    {-8,26}
};
#else
/*#if defined(APA16) || defined(MIPSCOLOR) || defined(MPEL) || defined(aiws)*/
private VECT asteroid[] = {
    {2,18},
    {10,22},
    {20,5},
    {17,-4},
    {7,-6},
    {16,-10},
    {12,-17},
    {-7,-19},
    {-17,-13},
    {-21,3},
    {-14,17},
    {-8,11},
    {-7,23}
};
#endif

/* if true, asteroids bounce off edge, otherwise they wrap */
static boolean bounce=FALSE;

void asteroids_GlobalSetBounce(b)
boolean b;
{
    bounce=b;
}

private int Scores[ASTSIZES];

private struct asteroid *AllocAsteroid(A)
ASTEROIDS *A;
{
    if (A->free == NULL)
        return New(struct asteroid);
    else {
	struct asteroid *new = A->free;
	A->free = new->next;
        return new;
    }
}

private void FreeAsteroid(A,a)
ASTEROIDS *A;
struct asteroid *a;
{
    sprite_Destroy(a->sprite);
    debris_Destroy(a->debris);
    a->next = A->free;
    A->free = a;
}

void asteroids_GlobalInit()
{
    int size, rot;
    VECT mag;

    Shapes[0][0] =
      shape_CreatePolygon(asteroid, sizeof(asteroid) / sizeof(asteroid[0]));
    vect_SetXY(&mag, 2, 1);
    shape_Scale(Shapes[0][0], &mag);

    for (rot = 1; rot < ASTROTATIONS; rot++) {
        Shapes[0][rot] = shape_Create();
        shape_Copy(Shapes[0][rot], Shapes[0][0]);
        shape_Rotate(Shapes[0][rot], 360 * rot / ASTROTATIONS);
    }

    vect_SetY(&mag, ASTSIZES);
    for (size = 1 ; size < ASTSIZES; size++) {
        vect_SetX(&mag, ASTSIZES-size);
        for (rot = 0; rot < ASTROTATIONS; rot++) {
            Shapes[size][rot] = shape_Create();
            shape_Copy(Shapes[size][rot], Shapes[0][rot]);
            shape_Scale(Shapes[size][rot], &mag);
        }
    }

    for (size = 0; size < ASTSIZES; size++)
        Scores[size] = MULASTSCORE * size + BASEASTSCORE;
}

private void NewAsteroid(A, size, pos, vel, inc)
ASTEROIDS *A;
int size;
VECT *pos;
VECT *vel;
int inc;
{
    VECT screen;
    struct asteroid *this = AllocAsteroid(A);

    plot_GetSize(&screen);

    this->size = size;
    this->whole = TRUE;
    this->sprite = sprite_Create();
    sprite_SetShapes(this->sprite, &Shapes[this->size][0], ASTROTATIONS);
    sprite_SetIndex(this->sprite, rand_Natural(ASTROTATIONS));
    sprite_SetInc(this->sprite, inc);
    sprite_SetPos(this->sprite, pos);
    sprite_SetVel(this->sprite, vel);
    sprite_SetBounds(this->sprite,
      1, vect_GetX(&screen) - 2, SCOREBARH + 1, vect_GetY(&screen) - 2);

    if(bounce)
	sprite_SetBounce(this->sprite,TRUE);
    else
	sprite_SetWrap(this->sprite,TRUE);

    this->debris = debris_Create(0,5,20);

    sprite_On(this->sprite);

    this->next = A->active;
    A->active=this;
}

/*
 * Pick an initial random position and velocity
 * 1/edge is the fraction of the screen size along the
 * edges where asteroids can initially occur.
 */

private void Randeroid(pos, vel, edge)
VECT *pos;
VECT *vel;
long edge;
{
    VECT size;
    long x, y;

    plot_GetSize(&size);
    vect_GetXY(&size, &x, &y);

    /* Pick a random side of the screen to start on */

    switch (rand_Natural(4)) {
       case 0:
           vect_SetXY(pos, rand_Natural(x),
             rand_Natural(y / edge));
           break;
       case 1:
           vect_SetXY(pos, rand_Natural(x),
             y * (edge - 1) / edge + rand_Natural(y / edge));
           break;
       case 2:
           vect_SetXY(pos, rand_Natural(y / edge),
             rand_Natural(y));
           break;
       default:
           vect_SetXY(pos, x * (edge - 1) / edge + rand_Natural(x / edge),
             rand_Natural(y));
           break;
    }

    vect_SetXY(vel, rand_Integer(ADJASTVEL), rand_Integer(ADJASTVEL));
}

void asteroids_Start(A)
ASTEROIDS *A;
{
    struct asteroid *next, *this;
    int i;

    for (this = A->active; this != NULL; this = next) {
        next = this->next;
        FreeAsteroid(A,this);
    }

    A->active=NULL;
    A->delay = -1;

    for (i = 0; i < A->wavesize; i++) {
        VECT pos, vel;
        Randeroid(&pos, &vel, 4);
        NewAsteroid(A, 0, &pos, &vel, rand_Integer(MAXASTROT));
    }
}

void asteroids_On(A)
ASTEROIDS *A;
{
    int flushcount = 0;
    struct asteroid *ptr;

    for (ptr = A->active; ptr != NULL; ptr = ptr->next) {
        if (ptr->whole)
            sprite_On(ptr->sprite);
        if (++flushcount == 20) {
            OPTIONAL_FLUSH();
            flushcount = 0;
        }
    }

    OPTIONAL_FLUSH();
}

void asteroids_Off(A)
ASTEROIDS *A;
{
    struct asteroid *ptr;
    int flushcount = 0;

    for (ptr = A->active; ptr != NULL; ptr = ptr->next) {
        if (ptr->whole)
            sprite_Off(ptr->sprite);
        if (++flushcount == 20) {
            OPTIONAL_FLUSH();
            flushcount = 0;
        }
    }

    OPTIONAL_FLUSH();
}

void asteroids_Update(A)
ASTEROIDS *A;
{
    struct asteroid *ptr, **prev;
    int flushcount = 0;

    if (A->active == NULL) {
        if (A->delay == -1)
            A->delay = WAVEDELAY;
        else if (A->delay > 0)
            A->delay--;
        else {
            A->wavesize += 2;
            asteroids_Start(A);
        }
    }       

    for (prev = &A->active; (ptr = *prev) != NULL; ) {
        if (ptr->whole) {
            sprite_Update(ptr->sprite);
            prev = &ptr->next;
        }else{
            debris_Update(ptr->debris);
            if (debris_IsClear(ptr->debris)) {
                *prev = ptr->next;
                FreeAsteroid(A, ptr);
            }
            else
                prev = &ptr->next;
        }
        if (++flushcount == 20) {
            OPTIONAL_FLUSH();
            flushcount = 0;
        }
    }

    OPTIONAL_FLUSH();
}

/* Returns number of points scored */

static int split(A, ast)
ASTEROIDS *A;
struct asteroid *ast;
{
    VECT vel, rnd, pos;
    int score = Scores[ast->size];
    
    sprite_GetPos(ast->sprite, &pos);

    if (ast->size < ASTSIZES - 1) {
        sprite_GetVel(ast->sprite, &vel);
        vect_SetXY(&rnd, rand_Integer(ADJASTVEL), rand_Integer(ADJASTVEL));
        vect_Add(&vel, &rnd);
        NewAsteroid(A, ast->size + 1, &pos, &vel,
                    sprite_GetInc(ast->sprite) + rand_Integer(MAXASTROT));
        sprite_GetVel(ast->sprite, &vel);
        vect_SetXY(&rnd, rand_Integer(ADJASTVEL), rand_Integer(ADJASTVEL));
        vect_Add(&vel, &rnd);
        NewAsteroid(A, ast->size + 1, &pos, &vel,
                    sprite_GetInc(ast->sprite) + rand_Integer(MAXASTROT));
    }
    else if (A->active == NULL)
        A->delay = WAVEDELAY;
    
    ast->whole = FALSE;
    sprite_Off(ast->sprite);
    sprite_Update(ast->sprite);
    sprite_GetVel(ast->sprite, &vel);
    debris_SetPos(ast->debris, &pos);
    debris_SetVel(ast->debris, &vel);
    debris_Start(ast->debris);

    soundeff_AstExplode();

    return score;
}

/*
 * Test collision of any asteroid with shape sh at position p.
 * Splits/destroys asteroids.
 *
 * Returns number of points scored.
 */

int asteroids_HitSprite(A, s)
ASTEROIDS *A;
SPRITE *s;
{
    struct asteroid *ptr;

    for (ptr = A->active; ptr != NULL; ptr = ptr->next) {
        if (!ptr->whole)
            continue;
        if (shape_Collision(sprite_GetCurShape(s), sprite_RefPos(s),
          sprite_GetCurShape(ptr->sprite), sprite_RefPos(ptr->sprite)))
            return split(A, ptr);
    }

    return 0;
}

/*
 * Test collision of any asteroid with a missile.
 * Splits/destroys asteroids.
 *
 * Returns number of points scored.
 */

int asteroids_HitMissile(A, m)
ASTEROIDS *A;
MISSILE *m;
{
    struct asteroid *ptr;

    for (ptr = A->active; ptr != NULL; ptr = ptr->next)
        if (!ptr->whole)
            continue;
        else if (missile_HitSprite(m, ptr->sprite))
            return split(A, ptr);

    return 0;
}

ASTEROIDS *asteroids_Create()
{
    ASTEROIDS *A = New(ASTEROIDS);
    A->active = NULL;
    A->free = NULL;
    A->wavesize = INITWAVESIZE;
    A->delay = 0;
    return A;
}

void asteroids_Destroy(A)
ASTEROIDS *A;
{
    struct asteroid *next, *this;

    for (this = A->active; this != NULL; this = next) {
        next = this->next;
        FreeAsteroid(A,this);
    }

   Zap(A);
}
