#include "game.h"
#include "ship.h"
#include "asteroids.h"
#include "control.h"
#include "ghost.h"
#include "text.h"
#include "frac.h"
#include "rand.h"
#include "soundeff.h"
#include "high.h"

private boolean soundIsOn = FALSE;
private boolean soundOverride = FALSE;

void game_GlobalInit()
{
    soundIsOn = FALSE;
    soundOverride = TRUE;
    sound_Off();
}

GAME *game_Create(player,level)
int player,level;
{
    GAME *g = New(GAME);
    VECT fontSize, size, pos;

    vect_SetXY(&fontSize, 2, 1);
    g->scoreFont = linefont_Create(&fontSize, 0);
    g->score = score_Create();
    g->scoreThreshold = EXTRAMANINTERVAL;
    g->ship=ship_Create();
    g->extras = extras_Create();
    extras_Set(g->extras, STARTLIVES);
    g->aster=asteroids_Create();
    g->updates=updlist_Create();
    g->player=player;
    g->level=level;
    g->alien=alien_Create();
    g->tst = game_WAIT;

    score_SetFont(g->score, g->scoreFont);
    score_GetSize(g->score, &size);

    vect_SetXY(&pos,SCOREX,(SCOREBARH - vect_GetY(&size)) / 2);
    score_SetPos(g->score, &pos);

    asteroids_Start(g->aster);

    return g;
}

void game_Destroy(g)
GAME *g;
{
    asteroids_Destroy(g->aster);
    score_Destroy(g->score);
    ship_Destroy(g->ship);
    updlist_Destroy(g->updates);
    extras_Destroy(g->extras);
    alien_Destroy(g->alien);
    linefont_Destroy(g->scoreFont);
    Zap(g);
}

struct missileCollRock {
    int score;
    ALIEN *alien;
    ASTEROIDS *aster;
};

static int missileColl(m, rock)
MISSILE *m;
struct missileCollRock *rock;
{
    int score;

    score = asteroids_HitMissile(rock->aster, m);

    if (score == 0 && alien_State(rock->alien) == alien_ALIVE)
        score = alien_HitMissile(rock->alien, m);

    if (score != 0) {
	missile_Off(m);
	rock->score += score;
    }

    return TRUE;
}

static void FindCollisions(g)
GAME *g;
{
    int score;
    struct missileCollRock rock;
    boolean died = FALSE;

    if (g->tst == game_ACTIVE ||
          g->tst == game_BOOM || g->tst == game_DEAD) {
	rock.score=0;
        rock.alien=g->alien;
	rock.aster=g->aster;
	mlist_Enumerate(ship_GetMissiles(g->ship),missileColl,&rock);

	score_Inc(g->score,rock.score);

        if (score_Get(g->score) >= g->scoreThreshold) {
            g->scoreThreshold += EXTRAMANINTERVAL;
            extras_Inc(g->extras);
            soundeff_ExtraMan();
        }
    }

    if (g->tst == game_ACTIVE) {
        if(!ship_IsInHyperspace(g->ship)){
            score=asteroids_HitSprite(g->aster, ship_GetSprite(g->ship));
            if(score!=0){
                score_Inc(g->score,score);
                died=TRUE;
            }
        }

        score=0;
        if(alien_State(g->alien) == alien_ALIVE &&
	   alien_CheckCollisions(g->alien,g->aster,g->ship,&score))
            died=TRUE;

        if(score!=0)
            score_Inc(g->score, score);

        if(died){
            ship_Die(g->ship);
            soundeff_ShipExplode();
	    g->tst=game_BOOM;
	}
    }

    if(alien_State(g->alien) == alien_ALIVE)
	alien_CheckCollisions(g->alien, g->aster, NULL);
}

static void Update(g)
GAME *g;
{
    asteroids_Update(g->aster);
    updlist_Update(g->updates);
    if(g->tst!=game_ATTRACT){
	score_Update(g->score);
	ship_Update(g->ship);
	if (g->tst==game_BOOM && !ship_IsExploding(g->ship))
            g->tst=game_DEAD;
    }

    if (alien_State(g->alien) == alien_GONE && g->tst != game_WAIT &&
	rand_Natural(ALIENCREATIONPROB) == 0)
        alien_Start(g->alien);

    alien_Update(g->alien, g->ship, g->level);

    FindCollisions(g);
}

static enum control_Command GetCommand(g)
GAME *g;
{
    boolean pause=FALSE;
    enum control_Command command;
    enum game_TURNSTATE initState=g->tst;

    do{
	command=control_NextCommand();

	switch(command){
	case control_PAUSE:
            if (g->tst != game_ATTRACT) {
                soundOverride = TRUE;
                sound_Off();
                pause=TRUE;
            }
	    break;
        case control_TOGGLESOUND:
            if (soundIsOn = ! soundIsOn) {
                if (! soundOverride)
                    sound_On();
            } else
                sound_Off();
            break;
	case control_TIMEUP:
	    control_BeginLoop();
	    if(!pause)
		Update(g);
	    break;
	case control_QUIT:
	    g->tst = game_QUIT;
	    break;
	default:
            if (pause) {
                pause=FALSE;
                soundOverride = FALSE;
                if (soundIsOn)
                    sound_On();
            }
        }
    } while ((command == control_TIMEUP && g->tst == initState) || pause);

    return command;
}

static int Start(g)
GAME *g;
{
#ifdef sun
    /* Sun3 screens tend to scroll.  We should probably trap and discard
     * console output, but I'm too lazy. --jgm
     */
    plot_ClearBlack();
#endif

    asteroids_On(g->aster);
    alien_On(g->alien);
    score_On(g->score);

    control_BeginLoop();

    return TRUE;
}

static void Stop(g)
GAME *g;
{
    ship_Off(g->ship);
    ship_Update(g->ship);
    score_Off(g->score);
    score_Update(g->score);
    asteroids_Off(g->aster);
    asteroids_Update(g->aster);
    extras_Off(g->extras);
    extras_Update(g->extras);
    alien_Update(g->alien, g->ship, g->level);
    alien_Off(g->alien);
}

/* Returns FALSE if quit is hit or out of men */

int game_Turn(g)
GAME *g;
{
    GHOST *ghost;
    enum control_Command command;

    if (extras_Get(g->extras) == 0)
        return FALSE;
    extras_Dec(g->extras);

    extras_On(g->extras);
    extras_Update(g->extras);
    updlist_Add(g->updates, g->extras, extras_Update);

    ghost = ghost_Create(g->player);

    g->tst=game_WAIT;
    Start(g);

    updlist_Add(g->updates,ghost,ghost_Update);

    if(GetCommand(g)!=control_QUIT)
	g->tst=game_ACTIVE;

    updlist_Remove(g->updates,ghost);
    ghost_Destroy(ghost);

    ship_Start(g->ship);

    while(g->tst==game_ACTIVE){
        command=GetCommand(g);
	if(g->tst==game_ACTIVE)
	    ship_Command(g->ship,command);
    }

    while(g->tst==game_BOOM)
	GetCommand(g);

    extras_Off(g->extras);
    extras_Update(g->extras);
    updlist_Remove(g->updates, g->extras);

    Stop(g);

    if (g->tst==game_QUIT)
	return FALSE;
    else
	return TRUE;
}

/* Returns the number of players */

int game_Attract()
{
    GAME *g=game_Create(0,0);
    enum control_Command command;
    TEXT *titleText = text_Create();
    LINEFONT *titleFont;
    TEXT *instrText = text_Create();
    LINEFONT *instrFont;
    VECT fontSize, textSize, screenSize, textPos;

    soundOverride = TRUE;
    sound_Off();

    text_Set(titleText,"ASTEROIDS");
    vect_SetXY(&fontSize,15,1);
    titleFont=linefont_Create(&fontSize,0);
    text_SetFont(titleText,titleFont);
    text_GetSize(titleText,&textSize);
    plot_GetSize(&screenSize);
    vect_SetXY(&textPos,
	       (vect_GetX(&screenSize)-vect_GetX(&textSize))/2,
	       (vect_GetY(&screenSize)-vect_GetY(&textSize))/2);
    text_SetPos(titleText, &textPos);
    text_On(titleText);

    text_Set(instrText,"Press 1 or 2 player Start");
    vect_SetXY(&fontSize,3,1);
    instrFont = linefont_Create(&fontSize, 0);
    text_SetFont(instrText, instrFont);
    text_GetSize(instrText, &textSize);
    vect_SetXY(&textPos,
	       (vect_GetX(&screenSize)-vect_GetX(&textSize))/2,
	       (vect_GetY(&textPos)-vect_GetY(&textSize)*3/2));
    text_SetPos(instrText, &textPos);
    text_On(instrText);

    {
        VECT hpos, hsize;
        vect_SetXY(&hpos, 0, vect_GetY(&screenSize) * 2 / 3);
        vect_SetXY(&hsize, vect_GetX(&screenSize), vect_GetY(&screenSize) / 3);
        high_Display(&hpos, &hsize);
    }

    game_AddUpdate(g,titleText,text_Update);
    game_AddUpdate(g,instrText,text_Update);
    game_AddUpdate(g,0,high_Update);

    g->tst=game_ATTRACT;
    Start(g);

    do
	command=GetCommand(g);
    while(command!=control_QUIT && command!=control_START &&
	  command!=control_START_2);

    Stop(g);
    game_Destroy(g);

    text_Destroy(titleText);
    text_Destroy(instrText);
    linefont_Destroy(titleFont);
    linefont_Destroy(instrFont);

    high_Undisplay();

    soundOverride = FALSE;
    if (soundIsOn)
        sound_On();

    if(g->tst==game_QUIT)
	return 0;
    else
	return (command == control_START_2) ? 2 : 1;
}
