/*
 * Input driver for DECstation 3100
 *
 * Note: Requires use of the kernel X driver because the DEC keyboard
 * driver has been very badly implemented.
 */

#include <sys/time.h>
#include <sys/file.h>
#include <mips/pmioctl.h>
#include <mips/pmevent.h>

#include "mips.keynames.h"
#include "type.h"
#include "input.h"
#include "plot.h"

static pmEventQueue *eventQueue;
static int fd;

int mips_InitializeInput(info)
    struct pm_info **info;
{

    static int initialized = FALSE;
    static int mouseFD;
    static struct pm_info *pmInfo;

    if (!initialized) {
        mouseFD = open("/dev/mouse", O_RDWR | O_NDELAY);

        if (ioctl(mouseFD, QIOCINIT, 0) < 0) {
            close(mouseFD);
            return -1;
        }

        if (ioctl(mouseFD, QIOCGINFO, &pmInfo) < 0) {
            close(mouseFD);
            return -1;
        }

        initialized = TRUE;
    }

    if (info != NULL)
        *info = pmInfo;

    return mouseFD;
}

boolean input_GlobalInit()
{
    struct pm_info *pmInfo;
    struct pm_kpcmd ioc; 
    int divset;

    fd = mips_InitializeInput(&pmInfo);
    if (fd < 0)
        return FALSE;

#define LK_UPDOWN 0x86

    for (divset = 1; divset < 15; divset++) {
            ioc.nbytes = 0;
            ioc.cmd = LK_UPDOWN | (divset << 3);
            if (ioctl(fd, QIOCKPCMD, &ioc) < 0)
                return FALSE; /* Might leave keyboard in a bad way, oh well. */
    }

    eventQueue = &pmInfo->qe;
    eventQueue->eHead = eventQueue->eTail;

    return TRUE;
}

void input_Done()
{
    close(fd);
}

enum input_Status input_Poll(key, delay)
enum input_Key *key;
struct timeval *delay;
{
    enum input_Status result;
    pmEvent *event;
    int h;
    long mask = 1 << fd;
    static unsigned char keyStates[32];
    int thisKey = 0;
    int discardEvent;

    plot_Flush(); /* make sure all output is visible */

    select(fd + 1, &mask, NULL, NULL, delay);

    if (eventQueue->eHead == eventQueue->eTail)
        return input_NONE;

    event = eventQueue->events + eventQueue->eHead;

#define LK_ALLUP 0xb3
    if (event->key == LK_ALLUP) {
        
        int byteIndex = 0;

        while (byteIndex < sizeof(keyStates) && thisKey == 0) {
            
            int bitIndex = 0;
            unsigned int thisByte = keyStates[byteIndex];

            while (thisByte) {
                if (thisByte & 1) {
                    thisKey = (byteIndex << 3) + bitIndex;
                    break;
                }
                else {
                    ++bitIndex;
                    thisByte >>= 1;
                }
            }
            ++byteIndex;
        }

        discardEvent = (thisKey == 0);
    }
    else {
        thisKey = event->key;
        discardEvent = 1;
    }

    if (event->device != KEYBOARD_DEVICE) {
        result = input_NONE;
    }
    else {
        if (keyStates[thisKey >> 3] & (1 << (thisKey % 8))) {
            result = input_KEYUP;
            keyStates[thisKey >> 3] &= !(1 << (thisKey % 8));
        }
        else {
            result = input_KEYDOWN;
            keyStates[thisKey >> 3] |= (1 << (thisKey % 8));
        }
     }

    switch (thisKey) {
      case KEY_LANGLE_RANGLE: /* <> */
        *key = key_ROTLEFT;
        break;

      case KEY_Q: /* Q */
        *key = key_QUIT;
        break;

      case KEY_PERIOD: /* Period */
        *key = key_THRUST;
        break;

      case KEY_TR_1: /* 1 */
      case KEY_KP_1: /* Num-1 */
        *key = key_START;
        break;

      case KEY_TR_2: /* 2 */
      case KEY_KP_2: /* Num-2 */
        *key = key_START_2;
        break;

      case KEY_Z: /* Z */
        *key = key_ROTRIGHT;
        break;

      case KEY_P: /* p */
        *key = key_PAUSE;
        break;

      case KEY_S: /* s */
        *key = key_TOGGLESOUND;
        break;

      case KEY_QMARK: /* ? */
        *key = key_FIRE;
        break;

      case KEY_SPACE: /* space bar */
        *key = key_HYPERSPACE;
        break;

      default:
        result = input_NONE;
        break;
    }

    if (discardEvent) {
        h = eventQueue->eHead + 1;
        if (h > eventQueue->eSize)
            h = 0;
        eventQueue->eHead = h;
    }

    return result;
}

