#include "shape.h"
#include "cordic.h"

#define INITIALSIZE 10
#define SQTABSIZE 100

#define ABS(a) ((a) < 0 ? -(a) : (a))

private long SqTab[SQTABSIZE];

void shape_GlobalInit()
{
    register int i = 0, s1 = 0, s2 = 1, s3 = 2;

    /* Compute table of squares using forward sums */

    while (i < SQTABSIZE)
        SqTab[i++] = s1, s1 += s2, s2 += s3;
}

SHAPE *shape_Create()
{
    SHAPE *s = New(SHAPE);

    s->count = 0;
    s->alloc = INITIALSIZE;
    s->segs = NewN(LINE, s->alloc);
    s->rad = 0;

    /*
     * The radius is computed when needed and is cached.
     */

    return s;
}

void shape_Destroy(s)
SHAPE *s;
{
    if (s->count)
        ZapN(s->segs);
    Zap(s);
}

void shape_Copy(s1, s2)
SHAPE *s1, *s2;
{
    LINE *lp = s2->segs;
    int i = s2->count;

    s1->count = 0;

    while (i--)
        shape_Insert(s1, lp++);

    s1->rad = s2->rad;
}

SHAPE *shape_CreatePolygon(verts, num_verts)
VECT *verts;
int num_verts;
{
    SHAPE *s = shape_Create();
    VECT *prev, *next;
    LINE l;

    prev = verts + num_verts - 1;
    next = verts;

    while (num_verts-- > 0) {
        line_SetFT(&l, prev, next);
        shape_Insert(s, &l);
        prev = next;
        next++;
    }

    return s;
}

void shape_Insert(s, l)
SHAPE *s;
LINE *l;
{
    if (s->count == s->alloc) {
        s->alloc <<= 1;
        s->segs = OldN(LINE, s->alloc, s->segs);
    }

    line_Copy(&s->segs[s->count++], l);
    s->rad = 0;     /* Invalidated */
}

/*
 * Enumeration: proc is called for each segment with
 * arguments:   (long rock, LINE *line)
 * `Rock' may be used for any purpose (via casts).
 * If proc ever returns FALSE, enumeration terminates
 * and returns FALSE, else enumeration returns TRUE.
 *
 * This routine used to be used in Draw but Draw
 * needs to be fast so it's expanded in Draw.
 */

boolean shape_Enumerate(s, proc, rock)
SHAPE *s;
boolean (*proc)();
long rock;
{
    LINE *lp = s->segs;
    register int i = s->count;

    while (i--)
        if (! (*proc)(lp++, rock))
            return FALSE;

    return TRUE;
}

void shape_Rotate(s, ang)
SHAPE *s;
long ang;
{
    LINE *sp;
    register int i = s->count;

    for (sp = s->segs; i--; sp++) {
        cordic_Rotate(line_RefF(sp), ang);
        cordic_Rotate(line_RefT(sp), ang);
    }
}

void shape_Scale(s, scale)
SHAPE *s;
FRAC *scale;
{
    LINE *sp;
    register int i = s->count;

    for (sp = s->segs; i--; sp++) {
        vect_Scale(line_RefF(sp), scale);
        vect_Scale(line_RefT(sp), scale);
    }

    s->rad = 0;     /* Invalidated */
}

/*
 * Faster integer square root test routine.
 * This is actually the same algorithm you learned in grade school and
 * have painfully rediscovered every five years since then.  Thanks to
 * Charles Mesztenyi for suggesting the binary analog.
 * Ben Cranston 4/23/87
 */

long isqrt(n)
long n;
{
    long a = 1, b = 1, c, as = 0, bs = 0;

    while (a <= n)
	a <<= 2, b <<= 1;

    while (b > 1 && n > 0) {
	a >>= 2, b >>= 1;
	c = n - (as | a);
	if (c >= 0) {
	    n = c;
	    as |= a << 1, bs |= b;
	}
	as >>= 1;
    }

    return bs;
}

long shape_GetRadius(s)
SHAPE *s;
{
    if (s->rad == 0) {
        LINE *lp;
        register int i = s->count;
        register long maxSqd = 0;
        for (lp = s->segs; i--; lp++) {
            register long sqd;
            sqd = vect_SumSqr(line_RefF(lp));
            if (sqd > maxSqd)
                maxSqd = sqd;
            sqd = vect_SumSqr(line_RefT(lp));
            if (sqd > maxSqd)
                maxSqd = sqd;
        }
        s->rad = isqrt(maxSqd);
    }
    return s->rad;
}

void shape_GetBox(s, min, max)
SHAPE *s;
VECT *min;
VECT *max;
{
    register int i = s->count;
    register LINE *lp = s->segs;
    int minX, minY, maxX, maxY;

    if (i == 0) {
        vect_SetXY(min, 0, 0);
        vect_SetXY(max, 0, 0);
        return;
    }

    vect_GetXY(line_RefF(lp), &minX, &minY);
    vect_GetXY(line_RefT(lp), &maxX, &maxY);

    while (i--) {
        int x, y;
        vect_GetXY(line_RefF(lp), &x, &y);
        if (x < minX)
            minX = x;
        if (x > maxX)
            maxX = x;
        if (y < minY)
            minY = y;
        if (y > maxY)
            maxY = y;
        vect_GetXY(line_RefT(lp), &x, &y);
        if (x < minX)
            minX = x;
        if (x > maxX)
            maxX = x;
        if (y < minY)
            minY = y;
        if (y > maxY)
            maxY = y;
        lp++;
    }

    vect_SetXY(min, minX, minY);
    vect_SetXY(max, maxX, maxY);
}    

/*
 * Computes the size of smallest enclosing box
 */

void shape_GetBoxSize(s, size)
SHAPE *s;
VECT *size;
{
    static VECT ones = {1, 1};
    VECT min;
    shape_GetBox(s, &min, size);
    vect_Sub(size, &min);
    vect_Add(size, &ones);
}

void shape_Draw(s, pos)
SHAPE *s;
VECT *pos;
{
    register LINE *lp;
    register int i;
    
    if (s == NULL)
        return;

    for (i = s->count, lp = s->segs; i--; lp++) {
        LINE ml;
        line_Copy(&ml, lp);
        line_AddVect(&ml, pos);
        plot_Move(line_RefF(&ml));
        plot_Line(line_RefT(&ml));
    }
}

/*
 * Expensively determine if two line segments intersect in any way
 * (22 32-bit multiplies, many more adds, subtracts and compares)
 */

private boolean Intersects(l0, l1)
LINE *l0, *l1;
{
    VECT *p0, *p1, *p2, *p3;
    long dx0, dy0, dx2, dy2;
    register long tmp, cmp1, cmp2, cmp3;

    p0 = line_RefF(l0);
    p1 = line_RefT(l0);
    p2 = line_RefF(l1);
    p3 = line_RefT(l1);

    dx0 = vect_GetX(p1) - vect_GetX(p0);
    dy0 = vect_GetY(p1) - vect_GetY(p0);
    dx2 = vect_GetX(p3) - vect_GetX(p2);
    dy2 = vect_GetY(p3) - vect_GetY(p2);

    tmp = dx2 * dy0 - dx0 * dy2;

    cmp2 = (vect_GetY(p2) - vect_GetY(p0)) * dx0 * dx2 +
           dy0 * dx2 * vect_GetX(p0) - dy2 * dx0 * vect_GetX(p2);

    cmp1 = vect_GetX(p0) * tmp;
    cmp3 = vect_GetX(p1) * tmp;
    if (! (cmp1 <= cmp2 && cmp2 <= cmp3 || 
           cmp3 <= cmp2 && cmp2 <= cmp1))
        return FALSE;

    cmp1 = vect_GetX(p2) * tmp;
    cmp3 = vect_GetX(p3) * tmp;
    if (! (cmp1 <= cmp2 && cmp2 <= cmp3 || 
           cmp3 <= cmp2 && cmp2 <= cmp1))
        return FALSE;

    cmp2 = (vect_GetX(p0) - vect_GetX(p2)) * dy0 * dy2 +
           dy0 * dx2 * vect_GetY(p2) - dy2 * dx0 * vect_GetY(p0);

    cmp1 = vect_GetY(p0) * tmp;
    cmp3 = vect_GetY(p1) * tmp;
    if (! (cmp1 <= cmp2 && cmp2 <= cmp3 || 
           cmp3 <= cmp2 && cmp2 <= cmp1))
        return FALSE;

    cmp1 = vect_GetY(p2) * tmp;
    cmp3 = vect_GetY(p3) * tmp;
    if (! (cmp1 <= cmp2 && cmp2 <= cmp3 || 
           cmp3 <= cmp2 && cmp2 <= cmp1))
        return FALSE;

    return TRUE;
}

/*
 * Returns TRUE if any of the line segments of one shape
 * are overlapping any of the line segments of another.
 * Given shapes and their positions.
 */

boolean shape_Collision(s0, p0, s1, p1)
SHAPE *s0;
VECT *p0;
SHAPE *s1;
VECT *p1;
{
    long dx, dy;
    long sumrad;
    long i, j, t;
    LINE l0, l1;
    VECT min, min0, min1, max, max0, max1;

    /* First a quick check is made for shape radii */

    dx = vect_GetX(p1) - vect_GetX(p0);
    dy = vect_GetY(p1) - vect_GetY(p0);
    dx = ABS(dx);
    dy = ABS(dy);

    sumrad = shape_GetRadius(s0) + shape_GetRadius(s1);

    if (dx >= SQTABSIZE || dy >= SQTABSIZE || sumrad >= SQTABSIZE)
        return FALSE;

    if (SqTab[dx] + SqTab[dy] > SqTab[sumrad])
        return FALSE;

    /* Slow M*N algorithm tests intersection of all line segment pairs */

    shape_GetBox(s0, &min0, &max0);
    vect_Add(&min0, p0);
    vect_Add(&max0, p0);

    shape_GetBox(s1, &min1, &max1);
    vect_Add(&min1, p1);
    vect_Add(&max1, p1);

    for (i = 0; i < s0->count; i++) {
        line_Copy(&l0, s0->segs + i);
        line_AddVect(&l0, p0);

        line_GetF(&l0, &min);
        line_GetT(&l0, &max);
        if (vect_GetX(&min) > vect_GetX(&max)) {
            t = vect_GetX(&min);
            vect_SetX(&min, vect_GetX(&max));
            vect_SetX(&max, t);
        }
        if (vect_GetY(&min) > vect_GetY(&max)) {
            t = vect_GetY(&min);
            vect_SetY(&min, vect_GetY(&max));
            vect_SetY(&max, t);
        }

        if (vect_GetX(&max) < vect_GetX(&min1))
            continue;
        if (vect_GetX(&min) > vect_GetX(&max1))
            continue;
        if (vect_GetY(&max) < vect_GetY(&min1))
            continue;
        if (vect_GetY(&min) > vect_GetY(&max1))
            continue;
        
        for (j = 0; j < s1->count; j++) {
            line_Copy(&l1, s1->segs + j);
            line_AddVect(&l1, p1);

            line_GetF(&l1, &min);
            line_GetT(&l1, &max);
            if (vect_GetX(&min) > vect_GetX(&max)) {
                t = vect_GetX(&min);
                vect_SetX(&min, vect_GetX(&max));
                vect_SetX(&max, t);
            }
            if (vect_GetY(&min) > vect_GetY(&max)) {
                t = vect_GetY(&min);
                vect_SetY(&min, vect_GetY(&max));
                vect_SetY(&max, t);
            }
                
            if (vect_GetX(&max) < vect_GetX(&min0))
                continue;
            if (vect_GetX(&min) > vect_GetX(&max0))
                continue;
            if (vect_GetY(&max) < vect_GetY(&min0))
                continue;
            if (vect_GetY(&min) > vect_GetY(&max0))
                continue;

            if (Intersects(&l0, &l1))
                return TRUE;
        }
    }

    return FALSE;
}

/*
 * shape_PointInside(sh, p)
 *
 * This routine is exact for any shape whose line segments
 * form a closed polygon, even if the line segments are not
 * in order.  It is a pretty good approximation if the shape
 * is not a closed polygon.
 *
 * The point is specified relative to the shape's origin.
 */

#define BETWEEN(a, b, tst) \
    ((tst) >= (a) && (tst) <= (b) || \
     (tst) >= (b) && (tst) <= (a))

struct erock {
    VECT *pt;
    long count;
};

private boolean EpiProc(l, erockp)
LINE *l;
struct erock *erockp;
{
    long tst;

    /* Check if Y of point is above or below extreme Y's of line */

    if (! BETWEEN(vect_GetY(line_RefF(l)),
      vect_GetY(line_RefT(l)), vect_GetY(erockp->pt)))
        return TRUE;    /* No intersections here */

    /* Case: Line is horizontal and point is on line */

    if (vect_GetY(line_RefF(l)) == vect_GetY(line_RefT(l)) &&
      BETWEEN(vect_GetX(line_RefF(l)),
      vect_GetX(line_RefT(l)), vect_GetX(erockp->pt)))
        return FALSE;

    /* Increment count only if interection of line & horiz. */
    /* line containing point is to the right of point. */

    tst = (vect_GetX(line_RefF(l)) - vect_GetX(line_RefT(l))) *
     (vect_GetY(erockp->pt) - vect_GetY(line_RefT(l))) -
     (vect_GetX(erockp->pt) - vect_GetX(line_RefT(l))) *
     (vect_GetY(line_RefF(l)) - vect_GetY(line_RefT(l)));

    if (vect_GetY(line_RefT(l)) > vect_GetY(line_RefF(l)) && tst > 0 ||
      vect_GetY(line_RefT(l)) < vect_GetY(line_RefF(l)) && tst < 0)
        erockp->count++;

    return TRUE;
}

boolean shape_PointInside(sh, pt)
SHAPE *sh;
VECT *pt;
{
    int rad, dx, dy;
    struct erock erock;

    dx = ABS(vect_GetX(pt));
    dy = ABS(vect_GetY(pt));

    /*
     * Two quick tests handle the very common trivial cases
     */

    if (dx > SQTABSIZE || dy > SQTABSIZE)
        return FALSE;

    rad = shape_GetRadius(sh);

    if (SqTab[dx] + SqTab[dy] > SqTab[rad])
        return FALSE;

    /*
     * Full test: count intersections with lines and a ray
     * extending from the point rightward to infinity.
     */

    erock.count = 0;
    erock.pt = pt;

    if (! shape_Enumerate(sh, EpiProc, (long) &erock))
        return TRUE;

    return (erock.count % 2);     /* TRUE if odd */
}

/*
 * The following routine tells if a given line is entirely or partially
 * touching a shape.  Note that the line is in the shape's coordinate
 * system.
 */

boolean shape_LineTouches(sh, l)
SHAPE *sh;
LINE *l;
{
    return (shape_PointInside(sh, line_RefF(l)) ||
            shape_PointInside(sh, line_RefT(l)));
}
